@@ -195,6 +195,26 @@ class ImageCollection(ABC):
         """
         pass
 
+    def apply_neighborhood(self,process, size:List[Dict],overlap:List[Dict]):
+        """
+        Applies a focal process to a data cube.
+        A focal process is a process that works on a 'neighbourhood' of pixels. The neighbourhood can extend into multiple dimensions, this extent is specified by the `size` argument. It is not only (part of) the size of the input window, but also the size of the output for a given position of the sliding window. The sliding window moves with multiples of `size`.
+
+        An overlap can be specified so that neighbourhoods can have overlapping boundaries. This allows for continuity of the output. The values included in the data cube as overlap can't be modified by the given `process`.
+
+        The neighbourhood size should be kept small enough, to avoid running beyond computational resources, but a too small size will result in a larger number of process invocations, which may slow down processing. Window sizes for spatial dimensions typically are in the range of 64 to 512 pixels, while overlaps of 8 to 32 pixels are common.
+
+        The process must not add new dimensions, or remove entire dimensions, but the result can have different dimension labels.
+
+        For the special case of 2D convolution, it is recommended to use ``apply_kernel()``.
+
+        @param process: Process to be applied on all neighbourhoods.
+        @param size: Neighbourhood sizes along each dimension. This object maps dimension names to either a physical measure (e.g. 100 m, 10 days) or pixels (e.g. 32 pixels). For dimensions not specified, the default is to provide all values. Be aware that including all values from overly large dimensions may not be processed at once.
+        @param overlap: Overlap of neighbourhoods along each dimension to avoid border effects. For instance a temporal dimension can add 1 month before and after a neighbourhood. In the spatial dimensions, this is often a number of pixels. The overlap specified is added before and after, so an overlap of 8 pixels will add 8 pixels on both sides of the window, so 16 in total. Be aware that large overlaps increase the need for computational resources and modifying overlapping data in subsequent operations have no effect.
+        @return: A data cube with the newly computed values. The cardinality, resolution and the number of dimensions are the same as for the original data cube.
+        """
+        pass
+
     def aggregate_time(self, temporal_window, aggregationfunction) -> 'ImageCollection' :
         """ Applies a windowed reduction to a timeseries by applying a user defined function.
             DEPRECATED: use Aggregate_temporal
@@ -206,7 +226,7 @@ class ImageCollection(ABC):
         """
         pass
 
-    def aggregate_temporal(self, intervals:List,labels:List, reducer, dimension:str = None) -> 'ImageCollection' :
+    def aggregate_temporal(self, intervals:List[List],reducer,labels:List = None, dimension:str = None,context:Dict=None) -> 'ImageCollection' :
         """ Computes a temporal aggregation based on an array of date and/or time intervals.
 
             Calendar hierarchies such as year, month, week etc. must be transformed into specific intervals by the clients. For each interval, all data along the dimension will be passed through the reducer. The computed values will be projected to the labels, so the number of labels and the number of intervals need to be equal.
@@ -222,6 +242,7 @@ class ImageCollection(ABC):
         """
         pass
 
+    @deprecated("Use a more specific reduce method instead.")
     def reduce(self,reducer,dimension):
         """
         Applies a reducer to a data cube dimension by collapsing all the input values along the specified dimension into a single output value computed by the reducer.
@@ -316,16 +337,6 @@ class ImageCollection(ABC):
         """
         pass
 
-    def stretch_colors(self, min, max) -> 'ImageCollection':
-        """ Color stretching
-
-            :param min: Minimum value
-            :param max: Maximum value
-
-            :return: An ImageCollection instance
-        """
-        pass
-
     def filter_bands(self, bands) -> 'ImageCollection':
         """Filters the bands in the data cube so that bands that don't match any of the criteria are dropped from the data cube.
         The data cube is expected to have only one spectral dimension.
@@ -361,12 +372,24 @@ class ImageCollection(ABC):
         """
         pass
 
-    def apply_kernel(self, kernel, factor=1.0) -> 'ImageCollection':
+    def apply_kernel(self, kernel, factor=1.0, border = 0, replace_invalid=0) -> 'ImageCollection':
         """
         Applies a focal operation based on a weighted kernel to each value of the specified dimensions in the data cube.
 
+        The border parameter determines how the data is extended when the kernel overlaps with the borders.
+        The following options are available:
+
+        * numeric value - fill with a user-defined constant number n: nnnnnn|abcdefgh|nnnnnn (default, with n = 0)
+        * replicate - repeat the value from the pixel at the border: aaaaaa|abcdefgh|hhhhhh
+        * reflect - mirror/reflect from the border: fedcba|abcdefgh|hgfedc
+        * reflect_pixel - mirror/reflect from the center of the pixel at the border: gfedcb|abcdefgh|gfedcb
+        * wrap - repeat/wrap the image: cdefgh|abcdefgh|abcdef
+
+
         :param kernel: The kernel to be applied on the data cube. The kernel has to be as many dimensions as the data cube has dimensions.
         :param factor: A factor that is multiplied to each value computed by the focal operation. This is basically a shortcut for explicitly multiplying each value by a factor afterwards, which is often required for some kernel-based algorithms such as the Gaussian blur.
+        :param border: Determines how the data is extended when the kernel overlaps with the borders. Defaults to fill the border with zeroes.
+        :param: replace_invalid: This parameter specifies the value to replace non-numerical or infinite numerical values with. By default, those values are replaced with zeroes.
         :return: A data cube with the newly computed values. The resolution, cardinality and the number of dimensions are the same as for the original data cube.
         """
         pass
@@ -383,6 +406,7 @@ class ImageCollection(ABC):
 
     ####VIEW methods #######
 
+    @deprecated(reason="use aggregate_spatial instead")
     def zonal_statistics(self, regions, func, scale=1000, interval="day") -> 'ImageCollection':
         """
         Calculates statistics for each zone specified in a file.
