@@ -5,30 +5,32 @@ This module provides a Connection object to manage and persist settings when int
 import logging
 import shutil
 import sys
+import warnings
 from pathlib import Path
 from typing import Dict, List, Tuple, Union, Callable
 from urllib.parse import urljoin
 
 import requests
 from deprecated import deprecated
-from openeo.internal.graph_building import PGNode
 from requests import Response
 from requests.auth import HTTPBasicAuth, AuthBase
 
 import openeo
 from openeo.capabilities import ApiVersionException, ComparableVersion
 from openeo.imagecollection import CollectionMetadata
+from openeo.internal.graph_building import PGNode
 from openeo.rest import OpenEoClientException
 from openeo.rest.auth.auth import NullAuth, BearerAuth
+from openeo.rest.auth.config import RefreshTokenStore, AuthConfig
 from openeo.rest.auth.oidc import OidcClientCredentialsAuthenticator, OidcAuthCodePkceAuthenticator, \
     OidcClientInfo, OidcAuthenticator, OidcRefreshTokenAuthenticator, OidcResourceOwnerPasswordAuthenticator, \
-    OidcDeviceAuthenticator, OidcProviderInfo, RefreshTokenStore
+    OidcDeviceAuthenticator, OidcProviderInfo
 from openeo.rest.datacube import DataCube
 from openeo.rest.imagecollectionclient import ImageCollectionClient
 from openeo.rest.job import RESTJob
-from openeo.rest.udp import RESTUserDefinedProcess, Parameter
 from openeo.rest.rest_capabilities import RESTCapabilities
-from openeo.util import ensure_list, get_user_data_dir
+from openeo.rest.udp import RESTUserDefinedProcess, Parameter
+from openeo.util import ensure_list, legacy_alias
 
 _log = logging.getLogger(__name__)
 
@@ -70,6 +72,10 @@ class RestApiConnection:
             )
         }
 
+    @property
+    def root_url(self):
+        return self._root_url
+
     def build_url(self, path: str):
         return url_join(self._root_url, path)
 
@@ -99,9 +105,10 @@ class RestApiConnection:
         )
         # Check for API errors and unexpected HTTP status codes as desired.
         status = resp.status_code
-        if check_error and status >= 400:
+        expected_status = ensure_list(expected_status) if expected_status else []
+        if check_error and status >= 400 and status not in expected_status:
             self._raise_api_error(resp)
-        if expected_status and status not in ensure_list(expected_status):
+        if expected_status and status not in expected_status:
             raise OpenEoClientException("Got status code {s!r} for `{m} {p}` (expected {e!r})".format(
                 m=method.upper(), p=path, s=status, e=expected_status)
             )
@@ -199,14 +206,21 @@ class Connection(RestApiConnection):
     # see https://github.com/Open-EO/openeo-wcps-driver/issues/45
     oidc_auth_user_id_token_as_bearer = False
 
-    def __init__(self, url, auth: AuthBase = None, session: requests.Session = None, default_timeout: int = None):
+    def __init__(
+            self, url, auth: AuthBase = None, session: requests.Session = None, default_timeout: int = None,
+            auth_config: AuthConfig = None, refresh_token_store: RefreshTokenStore = None
+    ):
         """
         Constructor of Connection, authenticates user.
 
         :param url: String Backend root url
         """
-        super().__init__(root_url=url, auth=auth, session=session, default_timeout=default_timeout)
-        self._cached_capabilities = None
+        self._orig_url = url
+        super().__init__(
+            root_url=self.version_discovery(url, session=session),
+            auth=auth, session=session, default_timeout=default_timeout
+        )
+        self._capabilities_cache = {}
 
         # Initial API version check.
         if self._api_version.below(self._MINIMUM_API_VERSION):
@@ -214,13 +228,48 @@ class Connection(RestApiConnection):
                 m=self._MINIMUM_API_VERSION, v=self._api_version)
             )
 
-    def authenticate_basic(self, username: str, password: str) -> 'Connection':
+        self._auth_config = auth_config
+        self._refresh_token_store = refresh_token_store or RefreshTokenStore()
+
+    @classmethod
+    def version_discovery(cls, url: str, session: requests.Session = None) -> str:
+        """
+        Do automatic openEO API version discovery from given url, using a "well-known URI" strategy.
+
+        :param url: initial backend url (not including "/.well-known/openeo")
+        :return: root url of highest supported backend version
+        """
+        try:
+            well_known_url_response = RestApiConnection(url, session=session).get("/.well-known/openeo")
+            assert well_known_url_response.status_code == 200
+            versions = well_known_url_response.json()["versions"]
+            supported_versions = [v for v in versions if cls._MINIMUM_API_VERSION <= v["api_version"]]
+            assert supported_versions
+            production_versions = [v for v in supported_versions if v.get("production", True)]
+            highest_version = max(production_versions or supported_versions, key=lambda v: v["api_version"])
+            _log.debug("Highest supported version available in backend: %s" % highest_version)
+            return highest_version['url']
+        except Exception:
+            # Be very lenient about failing on the well-known URI strategy.
+            return url
+
+    def _get_auth_config(self) -> AuthConfig:
+        if self._auth_config is None:
+            self._auth_config = AuthConfig()
+        return self._auth_config
+
+    def authenticate_basic(self, username: str = None, password: str = None) -> 'Connection':
         """
         Authenticate a user to the backend using basic username and password.
 
         :param username: User name
         :param password: User passphrase
         """
+        if username is None:
+            username, password = self._get_auth_config().get_basic_auth(backend=self._orig_url)
+            if username is None:
+                raise OpenEoClientException("No username/password given or found.")
+
         resp = self.get(
             '/credentials/basic',
             # /credentials/basic is the only endpoint that expects a Basic HTTP auth
@@ -294,22 +343,49 @@ class Connection(RestApiConnection):
             provider = OidcProviderInfo(discovery_url=self.build_url('/credentials/oidc'))
         return provider_id, provider
 
-    def _authenticate_oidc(self, authenticator: OidcAuthenticator, provider_id: str) -> 'Connection':
+    def _get_oidc_provider_and_client_info(
+            self, provider_id: str,
+            client_id: Union[str, None], client_secret: Union[str, None]
+    ) -> Tuple[str, OidcClientInfo]:
+        """
+        Resolve provider_id and client info (as given or from config)
+
+        :param provider_id: id of OIDC provider as specified by backend (/credentials/oidc).
+            Can be None if there is just one provider.
+
+        :return: (client_id, client_secret)
+        """
+        provider_id, provider = self._get_oidc_provider(provider_id)
+
+        if client_id is None:
+            client_id, client_secret = self._get_auth_config().get_oidc_client_configs(
+                backend=self._orig_url, provider_id=provider_id
+            )
+            _log.info("Using client_id {c!r} from config (provider {p!r})".format(c=client_id, p=provider_id))
+            if client_id is None:
+                raise OpenEoClientException("No client ID found.")
+
+        client_info = OidcClientInfo(client_id=client_id, client_secret=client_secret, provider=provider)
+
+        return provider_id, client_info
+
+    def _authenticate_oidc(
+            self,
+            authenticator: OidcAuthenticator,
+            provider_id: str,
+            store_refresh_token: bool = False
+    ) -> 'Connection':
         """
         Authenticate through OIDC and set up bearer token (based on OIDC access_token) for further requests.
         """
         tokens = authenticator.get_tokens()
         _log.info("Obtained tokens: {t}".format(t=[k for k, v in tokens._asdict().items() if v]))
-        if tokens.refresh_token:
-            # TODO: option to not store refresh token? Or opt-in to store refresh token?
-            try:
-                RefreshTokenStore().set(
-                    issuer=authenticator._client_info.provider.issuer,
-                    client_id=authenticator.client_id,
-                    refresh_token=tokens.refresh_token
-                )
-            except Exception as e:
-                _log.warning("Failed to store refresh token: {e!r}".format(e=e))
+        if tokens.refresh_token and store_refresh_token:
+            self._refresh_token_store.set_refresh_token(
+                issuer=authenticator.provider_info.issuer,
+                client_id=authenticator.client_id,
+                refresh_token=tokens.refresh_token
+            )
         token = tokens.access_token if not self.oidc_auth_user_id_token_as_bearer else tokens.id_token
         if self._api_version.at_least("1.0.0"):
             self.auth = BearerAuth(bearer='oidc/{p}/{t}'.format(p=provider_id, t=token))
@@ -319,92 +395,106 @@ class Connection(RestApiConnection):
 
     def authenticate_oidc_authorization_code(
             self,
-            client_id: str,
+            client_id: str = None,
             client_secret: str = None,
             provider_id: str = None,
             timeout: int = None,
             server_address: Tuple[str, int] = None,
-            webbrowser_open: Callable = None
+            webbrowser_open: Callable = None,
+            store_refresh_token=False,
     ) -> 'Connection':
         """
         OpenID Connect Authorization Code Flow (with PKCE).
 
         WARNING: this API is in experimental phase
         """
-        provider_id, provider = self._get_oidc_provider(provider_id)
-        # TODO: load client info and settings from config file?
-        client_info = OidcClientInfo(client_id=client_id, client_secret=client_secret, provider=provider)
+        provider_id, client_info = self._get_oidc_provider_and_client_info(
+            provider_id=provider_id, client_id=client_id, client_secret=client_secret
+        )
         authenticator = OidcAuthCodePkceAuthenticator(
             client_info=client_info,
             webbrowser_open=webbrowser_open, timeout=timeout, server_address=server_address
         )
-        return self._authenticate_oidc(authenticator, provider_id=provider_id)
+        return self._authenticate_oidc(authenticator, provider_id=provider_id, store_refresh_token=store_refresh_token)
 
     def authenticate_oidc_client_credentials(
             self,
-            client_id: str,
+            client_id: str = None,
             client_secret: str = None,
             provider_id: str = None,
+            store_refresh_token=False,
     ) -> 'Connection':
         """
         OpenID Connect Client Credentials flow.
 
         WARNING: this API is in experimental phase
         """
-        provider_id, provider = self._get_oidc_provider(provider_id)
-        # TODO: load credentials from file/config
-        client_info = OidcClientInfo(client_id=client_id, provider=provider, client_secret=client_secret)
+        provider_id, client_info = self._get_oidc_provider_and_client_info(
+            provider_id=provider_id, client_id=client_id, client_secret=client_secret
+        )
         authenticator = OidcClientCredentialsAuthenticator(client_info=client_info)
-        return self._authenticate_oidc(authenticator, provider_id=provider_id)
+        return self._authenticate_oidc(authenticator, provider_id=provider_id, store_refresh_token=store_refresh_token)
 
     def authenticate_oidc_resource_owner_password_credentials(
-            self, client_id: str, username: str, password: str, client_secret: str = None, provider_id: str = None
+            self,
+            username: str, password: str,
+            client_id: str = None,
+            client_secret: str = None,
+            provider_id: str = None,
+            store_refresh_token=False
     ) -> 'Connection':
         """
         OpenId Connect Resource Owner Password Credentials
 
         WARNING: this API is in experimental phase
         """
-        provider_id, provider = self._get_oidc_provider(provider_id)
-        # TODO: load password from file/config
-        client_info = OidcClientInfo(client_id=client_id, provider=provider, client_secret=client_secret)
+        provider_id, client_info = self._get_oidc_provider_and_client_info(
+            provider_id=provider_id, client_id=client_id, client_secret=client_secret
+        )
+        # TODO: also get username and password from config?
         authenticator = OidcResourceOwnerPasswordAuthenticator(
             client_info=client_info, username=username, password=password
         )
-        return self._authenticate_oidc(authenticator, provider_id=provider_id)
+        return self._authenticate_oidc(authenticator, provider_id=provider_id, store_refresh_token=store_refresh_token)
 
     def authenticate_oidc_refresh_token(
-            self, client_id: str, refresh_token: str = None, client_secret: str = None, provider_id: str = None
+            self, client_id: str = None, refresh_token: str = None, client_secret: str = None, provider_id: str = None
     ) -> 'Connection':
         """
         OpenId Connect Refresh Token
 
         WARNING: this API is in experimental phase
         """
-        provider_id, provider = self._get_oidc_provider(provider_id)
+        provider_id, client_info = self._get_oidc_provider_and_client_info(
+            provider_id=provider_id, client_id=client_id, client_secret=client_secret
+        )
+
         if refresh_token is None:
-            store = RefreshTokenStore()
-            # TODO: allow client_id/secret to be None and fetch it from config/cache?
-            refresh_token = store.get(issuer=provider.issuer, client_id=client_id)
+            refresh_token = self._refresh_token_store.get_refresh_token(
+                issuer=client_info.provider.issuer,
+                client_id=client_info.client_id
+            )
             if refresh_token is None:
-                raise OpenEoClientException("No refresh token")
+                raise OpenEoClientException("No refresh token given or found")
 
-        client_info = OidcClientInfo(client_id=client_id, provider=provider, client_secret=client_secret)
         authenticator = OidcRefreshTokenAuthenticator(client_info=client_info, refresh_token=refresh_token)
         return self._authenticate_oidc(authenticator, provider_id=provider_id)
 
     def authenticate_oidc_device(
-            self, client_id: str, client_secret: str, provider_id: str = None, **kwargs
+            self, client_id: str=None, client_secret: str=None, provider_id: str = None,
+            store_refresh_token=False,
+            **kwargs
     ) -> 'Connection':
         """
         Authenticate with OAuth Device Authorization grant/flow
 
         WARNING: this API is in experimental phase
         """
-        provider_id, provider = self._get_oidc_provider(provider_id)
-        client_info = OidcClientInfo(client_id=client_id, provider=provider, client_secret=client_secret)
+        provider_id, client_info = self._get_oidc_provider_and_client_info(
+            provider_id=provider_id, client_id=client_id, client_secret=client_secret
+        )
         authenticator = OidcDeviceAuthenticator(client_info=client_info, **kwargs)
-        return self._authenticate_oidc(authenticator, provider_id=provider_id)
+        return self._authenticate_oidc(authenticator, provider_id=provider_id, store_refresh_token=store_refresh_token)
 
     def describe_account(self) -> str:
         """
@@ -443,32 +533,11 @@ class Connection(RestApiConnection):
 
         :return: data_dict: Dict All available data types
         """
-        if self._cached_capabilities is None:
-            self._version_discovery()
-            base_url_response = self.get('/', check_error=False)
-            self._cached_capabilities = RESTCapabilities(base_url_response.json())
+        if "capabilities" not in self._capabilities_cache:
+            self._capabilities_cache["capabilities"] = RESTCapabilities(self.get('/').json())
+        return self._capabilities_cache["capabilities"]
 
-        return self._cached_capabilities
 
-    def _version_discovery(self):
-        try:
-            well_known_url_response = self.get('/.well-known/openeo')
-            assert well_known_url_response.status_code == 200
-            versions = well_known_url_response.json()["versions"]
-        except Exception:
-            # be very lenient about failing on the well-known doc, capabilities retrieval will be tried
-            return
-        supported_versions = [v for v in versions if self._MINIMUM_API_VERSION <= v["api_version"]]
-        if not supported_versions:
-            return
-        production_versions = [v for v in supported_versions if v.get("production", True)]
-        highest_version = max(production_versions or supported_versions, key=lambda v: v["api_version"])
-        _log.debug("Highest supported version available in backend: %s" % highest_version)
-        self._root_url = highest_version['url']
-
-    @deprecated("Use 'list_output_formats' instead")
-    def list_file_types(self) -> dict:
-        return self.list_output_formats()
 
     def list_output_formats(self) -> dict:
         if self._api_version.at_least("1.0.0"):
@@ -476,11 +545,15 @@ class Connection(RestApiConnection):
         else:
             return self.get('/output_formats').json()
 
+    list_file_types = legacy_alias(list_output_formats, "list_file_types")
+
     def list_file_formats(self) -> dict:
         """
         Get available input and output formats
         """
-        return self.get('/file_formats').json()
+        if "file_formats" not in self._capabilities_cache:
+            self._capabilities_cache["file_formats"] = self.get('/file_formats').json()
+        return self._capabilities_cache["file_formats"]
 
     def list_service_types(self) -> dict:
         """
@@ -512,7 +585,7 @@ class Connection(RestApiConnection):
     def collection_metadata(self, name) -> CollectionMetadata:
         return CollectionMetadata(metadata=self.describe_collection(name))
 
-    def list_processes(self) -> dict:
+    def list_processes(self) -> List[dict]:
         # TODO: Maybe format the result dictionary so that the process_id is the key of the dictionary.
         """
         Loads all available processes of the back end.
@@ -543,6 +616,9 @@ class Connection(RestApiConnection):
         :param public: visible to other users?
         :return: a RESTUserDefinedProcess instance
         """
+        if user_defined_process_id in set(p["id"] for p in self.list_processes()):
+            warnings.warn("Defining user-defined process {u!r} with same id as a pre-defined process".format(
+                u=user_defined_process_id))
         udp = RESTUserDefinedProcess(user_defined_process_id=user_defined_process_id, connection=self)
         udp.store(process_graph=process_graph, parameters=parameters, public=public)
         return udp
@@ -575,7 +651,7 @@ class Connection(RestApiConnection):
         # TODO make this a public property (it's also useful outside the Connection class)
         return self.capabilities().api_version_check
 
-    def datacube_from_process(self,process_id:str, **kwargs) -> DataCube:
+    def datacube_from_process(self, process_id: str, **kwargs) -> DataCube:
         """
         Load a raster datacube, from a custom process.
 
@@ -585,10 +661,11 @@ class Connection(RestApiConnection):
         """
 
         if self._api_version.at_least("1.0.0"):
-            graph = PGNode(process_id,kwargs)
-            return DataCube(graph,self)
+            graph = PGNode(process_id, kwargs)
+            return DataCube(graph, self)
         else:
-            raise OpenEoClientException("This method requires support for at least version 1.0.0 in the openEO backend.")
+            raise OpenEoClientException(
+                "This method requires support for at least version 1.0.0 in the openEO backend.")
 
     def load_collection(self, collection_id: str, **kwargs) -> Union[ImageCollectionClient, DataCube]:
         """
@@ -665,9 +742,10 @@ class Connection(RestApiConnection):
         return result
 
     # TODO: Maybe rename to execute and merge with execute().
-    def download(self, graph: dict, outputfile):
+    def download(self, graph: dict, outputfile: Union[Path, str, None] = None):
         """
-        Downloads the result of a process graph synchronously, and save the result to the given file.
+        Downloads the result of a process graph synchronously,
+        and save the result to the given file or return bytes object if no outputfile is specified.
         This method is useful to export binary content such as images. For json content, the execute method is recommended.
 
         :param graph: (flat) dict representing a process graph
@@ -675,8 +753,11 @@ class Connection(RestApiConnection):
         """
         request = self._build_request_with_process_graph(process_graph=graph)
         r = self.post(path="/result", json=request, stream=True, timeout=1000)
-        with Path(outputfile).open(mode="wb") as f:
-            shutil.copyfileobj(r.raw, f)
+        if outputfile is not None:
+            with Path(outputfile).open(mode="wb") as f:
+                shutil.copyfileobj(r.raw, f)
+        else:
+            return r.content
 
     def execute(self, process_graph: dict):
         """
@@ -732,14 +813,6 @@ class Connection(RestApiConnection):
         """
         return RESTJob(job_id, self)
 
-    def get_outputformats(self) -> dict:
-        """
-        Loads all available output formats.
-
-        :return: data_dict: Dict All available output formats
-        """
-        raise NotImplementedError()
-
     def load_disk_collection(self, format: str, glob_pattern: str, options: dict = {}) -> ImageCollectionClient:
         """
         Loads image data from disk as an ImageCollection.
@@ -749,7 +822,11 @@ class Connection(RestApiConnection):
         :param options: options specific to the file format
         :return: the data as an ImageCollection
         """
-        return ImageCollectionClient.load_disk_collection(self, format, glob_pattern, **options)
+
+        if self._api_version.at_least("1.0.0"):
+            return DataCube.load_disk_collection(self, format, glob_pattern, **options)
+        else:
+            return ImageCollectionClient.load_disk_collection(self, format, glob_pattern, **options)
 
 
 def connect(url, auth_type: str = None, auth_options: dict = {}, session: requests.Session = None,
