@@ -1,5 +1,6 @@
 """
-
+The main module for creating earth observation processes. It aims to easily build complex process chains, that can
+be evaluated by an openEO backend.
 
 .. data:: THIS
 
@@ -7,23 +8,30 @@
 
 """
 import datetime
+import inspect
 import logging
 import pathlib
 import typing
+from builtins import staticmethod
 from typing import List, Dict, Union, Tuple
 
+import numpy
 import numpy as np
 import shapely.geometry
 import shapely.geometry.base
 from deprecated import deprecated
 from shapely.geometry import Polygon, MultiPolygon, mapping
 
+import openeo
+import openeo.processes
 from openeo.imagecollection import ImageCollection, CollectionMetadata
 from openeo.internal.graph_building import PGNode, ReduceNode
-from openeo.rest import BandMathException, OperatorException
+from openeo.metadata import Band
+from openeo.processes import ProcessBuilder
+from openeo.rest import BandMathException, OperatorException, OpenEoClientException
 from openeo.rest.job import RESTJob
 from openeo.rest.udp import RESTUserDefinedProcess
-from openeo.util import get_temporal_extent, dict_no_none
+from openeo.util import get_temporal_extent, dict_no_none, legacy_alias
 from openeo.vectorcube import VectorCube
 
 if hasattr(typing, 'TYPE_CHECKING') and typing.TYPE_CHECKING:
@@ -39,13 +47,14 @@ THIS = object()
 
 class DataCube(ImageCollection):
     """
-    Class representing a OpenEO Data Cube.
+    Class representing a openEO Data Cube. Data loaded from the backend is returned as an object of this class.
+    Various processing methods can be invoked to build a complete workflow.
 
     Supports openEO API 1.0.
     In earlier versions this was called `ImageCollectionClient`
     """
 
-    def __init__(self, graph: PGNode, connection: 'Connection', metadata: CollectionMetadata = None):
+    def __init__(self, graph: PGNode, connection: 'openeo.Connection', metadata: CollectionMetadata = None):
         super().__init__(metadata=metadata)
         # Process graph
         self._pg = graph
@@ -69,7 +78,7 @@ class DataCube(ImageCollection):
         return self._connection.capabilities().api_version_check
 
     @property
-    def connection(self):
+    def connection(self) -> 'openeo.Connection':
         return self._connection
 
     def process(self, process_id: str, arguments: dict = None, metadata: CollectionMetadata = None, **kwargs) -> 'DataCube':
@@ -91,8 +100,7 @@ class DataCube(ImageCollection):
             arguments=arguments,
         ), metadata=metadata)
 
-    # Legacy `graph_add_node` method
-    graph_add_node = deprecated(reason="just use `process()`")(process)
+    graph_add_node = legacy_alias(process, "graph_add_node")
 
     def process_with_node(self, pg: PGNode, metadata: CollectionMetadata = None) -> 'DataCube':
         """
@@ -108,12 +116,12 @@ class DataCube(ImageCollection):
 
     @classmethod
     def load_collection(
-            cls, collection_id: str, connection: 'Connection' = None,
+            cls, collection_id: str, connection: 'openeo.Connection' = None,
             spatial_extent: Union[Dict[str, float], None] = None,
             temporal_extent: Union[List[Union[str, datetime.datetime, datetime.date]], None] = None,
             bands: Union[List[str], None] = None,
             fetch_metadata=True,
-            properties: Dict[str, PGNode] = None
+            properties: Dict[str, Union[str, PGNode, typing.Callable]] = None
     ):
         """
         Create a new Raster Data cube.
@@ -142,7 +150,7 @@ class DataCube(ImageCollection):
             arguments['bands'] = bands
         if properties:
             arguments['properties'] = {
-                prop: PGNode.to_process_graph_argument(pred)
+                prop: cls._get_callback(pred, parent_parameters=["value"])
                 for prop, pred in properties.items()
             }
         pg = PGNode(
@@ -153,13 +161,10 @@ class DataCube(ImageCollection):
             metadata = metadata.filter_bands(bands)
         return cls(graph=pg, connection=connection, metadata=metadata)
 
-    @classmethod
-    @deprecated("use load_collection instead")
-    def create_collection(cls, *args, **kwargs):
-        return cls.load_collection(*args, **kwargs)
+    create_collection = legacy_alias(load_collection, name="create_collection")
 
     @classmethod
-    def load_disk_collection(cls, connection: 'Connection', file_format: str, glob_pattern: str,
+    def load_disk_collection(cls, connection: 'openeo.Connection', file_format: str, glob_pattern: str,
                              **options) -> 'DataCube':
         """
         Loads image data from disk as a DataCube.
@@ -178,7 +183,7 @@ class DataCube(ImageCollection):
                 'options': options
             }
         )
-        return cls(graph=pg, connection=connection, metadata={})
+        return cls(graph=pg, connection=connection, metadata=CollectionMetadata({}))
 
     def _filter_temporal(self, start: str, end: str) -> 'DataCube':
         return self.process(
@@ -221,9 +226,7 @@ class DataCube(ImageCollection):
             cube.metadata = cube.metadata.filter_bands(bands)
         return cube
 
-    @deprecated("use `filter_bands()` instead")
-    def band_filter(self, bands) -> 'DataCube':
-        return self.filter_bands(bands)
+    band_filter = legacy_alias(filter_bands, "band_filter")
 
     def band(self, band: Union[str, int]) -> 'DataCube':
         """Filter the imagery by the given bands
@@ -340,6 +343,9 @@ class DataCube(ImageCollection):
         """
         return self._operator_binary("gt", other)
 
+    def __ge__(self, other: Union['DataCube', int, float]) -> 'DataCube':
+        return self._operator_binary("gte", other)
+
     def __lt__(self, other: Union['DataCube', int, float]) -> 'DataCube':
         """
         Pairwise comparison of the bands in this data cube with the bands in the 'other' data cube.
@@ -350,8 +356,8 @@ class DataCube(ImageCollection):
         """
         return self._operator_binary("lt", other)
 
-    def __truediv__(self, other) -> 'DataCube':
-        return self.divide(other)
+    def __le__(self, other: Union['DataCube', int, float]) -> 'DataCube':
+        return self._operator_binary("lte", other)
 
     def __add__(self, other) -> 'DataCube':
         return self.add(other)
@@ -374,6 +380,9 @@ class DataCube(ImageCollection):
     def __rmul__(self, other) -> 'DataCube':
         return self.multiply(other, reverse=True)
 
+    def __truediv__(self, other) -> 'DataCube':
+        return self.divide(other)
+
     def __rpow__(self, other) -> 'DataCube':
         return self.power(other,reverse=True)
 
@@ -452,11 +461,12 @@ class DataCube(ImageCollection):
         return self.merge(other, overlap_resolver=PGNode(
             process_id=operator,
             arguments={
-                left_arg_name: {"from_parameter": "cube1"},
-                right_arg_name: {"from_parameter": "cube2"},
+                left_arg_name: {"from_parameter": "x"},
+                right_arg_name: {"from_parameter": "y"},
             }
         ))
 
+    @deprecated(reason="use aggregate_spatial instead")
     def zonal_statistics(self, regions, func, scale=1000, interval="day") -> 'DataCube':
         """
         Calculates statistics for each zone specified in a file.
@@ -487,13 +497,67 @@ class DataCube(ImageCollection):
 
         return self.process(process_id, args)
 
+    @staticmethod
+    def _get_callback(process: Union[str, PGNode, typing.Callable], parent_parameters: List[str]) -> dict:
+        """
+        Build a "callback" process: a user defined process that is used by another process (such
+        as `apply`, `apply_dimension`, `reduce`, ....)
+
+        :param process: process id string, PGNode or callable that uses the ProcessBuilder mechanism to build a process
+        :parameter parameter_mapping: mapping of child (callback) parameters names to parent process parameter names
+        :return:
+        """
+
+        def get_parameter_names(process: typing.Callable) -> List[str]:
+            signature = inspect.signature(process)
+            return [
+                p.name for p in signature.parameters.values()
+                if p.kind in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
+            ]
+
+        # TODO: autodetect the parameters defined by process?
+        if isinstance(process, PGNode):
+            # Assume this is already a valid callback process
+            pg = process
+        elif isinstance(process, str):
+            # Assume given reducer is a simple predefined reduce process_id
+            if process in openeo.processes.__dict__:
+                process_params = get_parameter_names(openeo.processes.__dict__[process])
+            else:
+                # Best effort guess
+                process_params = parent_parameters
+            if parent_parameters == ["x", "y"] and (len(process_params) == 1 or process_params[:1] == ["data"]):
+                # Special case: wrap all parent parameters in an array
+                arguments = {process_params[0]: [{"from_parameter": p} for p in parent_parameters]}
+            else:
+                arguments = {a: {"from_parameter": b} for a, b in zip(process_params, parent_parameters)}
+            pg = PGNode(process_id=process, arguments=arguments)
+        elif isinstance(process, typing.Callable):
+            process_params = get_parameter_names(process)
+            if parent_parameters == ["x", "y"] and (len(process_params) == 1 or process_params[:1] == ["data"]):
+                # Special case: wrap all parent parameters in an array
+                arguments = [ProcessBuilder([{"from_parameter": p} for p in parent_parameters])]
+            else:
+                arguments = [ProcessBuilder({"from_parameter": p}) for p in parent_parameters]
+
+            pg = process(*arguments).pgnode
+        else:
+            raise ValueError(process)
+
+        return PGNode.to_process_graph_argument(pg)
+
     def apply_dimension(
-            self, code: str, runtime=None, version="latest", dimension='t', target_dimension=None
+            self, code: str = None, runtime=None,
+            process: [str, PGNode, typing.Callable] = None,
+            version="latest", dimension='t', target_dimension=None
     ) -> 'DataCube':
         """
-        Applies a user defined process to all pixel values along a dimension of a raster data cube. For example,
+        Applies a process to all pixel values along a dimension of a raster data cube. For example,
         if the temporal dimension is specified the process will work on a time series of pixel values.
 
+        The process to apply is specified by either `code` and `runtime` in case of a UDF, or by providing a callback function
+        in the `process` argument.
+
         The process reduce_dimension also applies a process to pixel values along a dimension, but drops
         the dimension afterwards. The process apply applies a process to each pixel value in the data cube.
 
@@ -506,8 +570,9 @@ class DataCube(ImageCollection):
         the dimension labels will be incrementing integers starting from zero, which can be changed using
         rename_labels afterwards. The number of labels will equal to the number of values computed by the process.
 
-        :param code: UDF code or process identifier
-        :param runtime: UDF runtime to use
+        :param code: UDF code or process identifier (optional)
+        :param runtime: UDF runtime to use (optional)
+        :param process: a callback function that creates a process graph, see :ref:`callbackfunctions`
         :param version: Version of the UDF runtime to use
         :param dimension: The name of the source dimension to apply the process on. Fails with a DimensionNotAvailable error if the specified dimension does not exist.
         :param target_dimension: The name of the target dimension or null (the default) to use the source dimension
@@ -518,35 +583,40 @@ class DataCube(ImageCollection):
         :raises: DimensionNotAvailable
         """
         if runtime:
-            process = self._create_run_udf(code, runtime, version)
+            # TODO EP-3555: unify better with UDF(PGNode) class and avoid doing same UDF code-runtime-version argument stuff in each method
+            callback_process_node = self._create_run_udf(code, runtime, version)
+            process = PGNode.to_process_graph_argument(callback_process_node)
+        elif code or process:
+            # TODO EP-3555 unify `code` and `process`
+            process = self._get_callback(code or process, parent_parameters=["data"])
         else:
-            process = PGNode(
-                process_id=code,
-                arguments={"data": {"from_parameter": "data"}},
-            )
+            raise OpenEoClientException("No UDF code or process given")
         arguments = {
-            "data": self._pg,
-            "process": PGNode.to_process_graph_argument(process),
+            "data": THIS,
+            "process": process,
             "dimension": self.metadata.assert_valid_dimension(dimension),
             # TODO #125 arguments: context
         }
         if target_dimension is not None:
             arguments["target_dimension"] = target_dimension
-        return self.process_with_node(PGNode(
-            process_id="apply_dimension",
-            arguments=arguments
-        ))
+        result_cube = self.process(process_id="apply_dimension", arguments=arguments)
 
-    def reduce_dimension(self, dimension: str, reducer: Union[PGNode, str],
-                         process_id="reduce_dimension", band_math_mode: bool = False) -> 'DataCube':
+        return result_cube
+
+    def reduce_dimension(
+            self, dimension: str, reducer: Union[str, PGNode, typing.Callable],
+            process_id="reduce_dimension", band_math_mode: bool = False
+    ) -> 'DataCube':
         """
         Add a reduce process with given reducer callback along given dimension
+
+        :param dimension: the label of the dimension to reduce
+        :param reducer: a callback function that creates a process graph, see :ref:`callbackfunctions`
         """
         # TODO: check if dimension is valid according to metadata? #116
         # TODO: #125 use/test case for `reduce_dimension_binary`?
-        if isinstance(reducer, str):
-            # Assume given reducer is a simple predefined reduce process_id
-            reducer = PGNode(process_id=reducer, arguments={"data": {"from_parameter": "data"}})
+        reducer = self._get_callback(reducer, parent_parameters=["data"])
+
         return self.process_with_node(ReduceNode(
             process_id=process_id,
             data=self._pg,
@@ -568,9 +638,22 @@ class DataCube(ImageCollection):
         """
         Apply reduce (`reduce_dimension`) process with given UDF along band/spectral dimension.
         """
+        # TODO EP-3555: unify better with UDF(PGNode) class and avoid doing same UDF code-runtime-version argument stuff in each method
         return self._reduce_bands(reducer=self._create_run_udf(code, runtime, version))
 
     def add_dimension(self, name: str, label: str, type: str = None):
+        """
+        Adds a new named dimension to the data cube.
+        Afterwards, the dimension can be referenced with the specified name. If a dimension with the specified name exists,
+        the process fails with a DimensionExists error. The dimension label of the dimension is set to the specified label.
+
+        This call does not modify the datacube in place, but returns a new datacube with the additional dimension.
+
+        @param name: The name of the dimension to add
+        @param label: The dimension label.
+        @param type: Dimension type, allowed values: 'spatial', 'temporal', 'bands', 'other', default value is 'other'
+        @return: The data cube with a newly added dimension. The new dimension has exactly one dimension label. All other dimensions remain unchanged.
+        """
         return self.process(
             process_id="add_dimension",
             arguments={"data": self._pg, "name": name, "label": label, "type": type},
@@ -578,7 +661,7 @@ class DataCube(ImageCollection):
         )
 
     def _create_run_udf(self, code, runtime, version) -> PGNode:
-        # TODO: expose this publicly (or create dedicated PGNode subclass)? Also encapsulate/decouple UDF loading better?
+        # TODO EP-3555: unify better with UDF(PGNode) class
         return PGNode(
             process_id="run_udf",
             arguments={
@@ -593,37 +676,62 @@ class DataCube(ImageCollection):
     def reduce_temporal_udf(self, code: str, runtime="Python", version="latest"):
         """
         Apply reduce (`reduce_dimension`) process with given UDF along temporal dimension.
-        """
-        return self._reduce_temporal(reducer=self._create_run_udf(code, runtime, version))
-
-    @deprecated("use `reduce_temporal_udf` instead")
-    def reduce_tiles_over_time(self, code: str, runtime="Python", version="latest"):
-        """
-        Applies a user defined function to a timeseries of tiles. The size of the tile is backend specific, and can be limited to one pixel.
-        The function should reduce the given timeseries into a single (multiband) tile.
 
         :param code: The UDF code, compatible with the given runtime and version
         :param runtime: The UDF runtime
         :param version: The UDF runtime version
+        """
+        # TODO EP-3555: unify better with UDF(PGNode) class and avoid doing same UDF code-runtime-version argument stuff in each method
+        return self._reduce_temporal(reducer=self._create_run_udf(code, runtime, version))
+
+    reduce_tiles_over_time = legacy_alias(reduce_temporal_udf, name="reduce_tiles_over_time")
+
+    def apply_neighborhood(
+            self, process: [str, PGNode, typing.Callable],
+            size: List[Dict], overlap: List[dict] = None
+    ) -> 'DataCube':
+        """
+        Applies a focal process to a data cube.
+
+        A focal process is a process that works on a 'neighbourhood' of pixels. The neighbourhood can extend into multiple dimensions, this extent is specified by the `size` argument. It is not only (part of) the size of the input window, but also the size of the output for a given position of the sliding window. The sliding window moves with multiples of `size`.
+
+        An overlap can be specified so that neighbourhoods can have overlapping boundaries. This allows for continuity of the output. The values included in the data cube as overlap can't be modified by the given `process`.
+
+        The neighbourhood size should be kept small enough, to avoid running beyond computational resources, but a too small size will result in a larger number of process invocations, which may slow down processing. Window sizes for spatial dimensions typically are in the range of 64 to 512 pixels, while overlaps of 8 to 32 pixels are common.\n\nThe process must not add new dimensions, or remove entire dimensions, but the result can have different dimension labels.
+
+        For the special case of 2D convolution, it is recommended to use ``apply_kernel()``.
+
+        :param size:
+        :param overlap:
+        :param process: a callback function that creates a process graph, see :ref:`callbackfunctions`
         :return:
         """
-        return self.reduce_temporal_udf(code=code, runtime=runtime, version=version)
-
-    def apply(self, process: Union[str, PGNode], data_argument='x') -> 'DataCube':
-        if isinstance(process, str):
-            # Simple single string process specification
-            process = PGNode(
-                process_id=process,
-                arguments={data_argument: {"from_parameter": "x"}}
+        return self.process(
+            process_id='apply_neighborhood',
+            arguments=dict_no_none(
+                data=THIS,
+                process=self._get_callback(process, parent_parameters=["data"]),
+                size=size,
+                overlap=overlap
             )
-        return self.process_with_node(PGNode(
-            process_id='apply',
+        )
+
+    def apply(self, process: Union[str, PGNode, typing.Callable] = None) -> 'DataCube':
+        """
+        Applies a unary process (a local operation) to each value of the specified or all dimensions in the data cube.
+
+        :param process: the name of a process, or a callback function that creates a process graph, see :ref:`callbackfunctions`
+        :param dimensions: The names of the dimensions to apply the process on. Defaults to an empty array so that all dimensions are used.
+        :return: A data cube with the newly computed values. The resolution, cardinality and the number of dimensions are the same as for the original data cube.
+        """
+        return self.process(
+            process_id="apply",
             arguments={
-                "data": self._pg,
-                "process": {"process_graph": process},
+                "data": THIS,
+                "process": self._get_callback(process, parent_parameters=["x"]),
                 # TODO #125 context
             }
-        ))
+        )
 
     def reduce_temporal_simple(self, process_id="max") -> 'DataCube':
         """Do temporal reduce with a simple given process as callback."""
@@ -668,6 +776,33 @@ class DataCube(ImageCollection):
         """
         return self.reduce_temporal_simple("count")
 
+    def aggregate_temporal(self, intervals:List[List],reducer: Union[str, PGNode, typing.Callable],labels:List = None, dimension:str = None, context:Dict=None) -> 'DataCube' :
+        """ Computes a temporal aggregation based on an array of date and/or time intervals.
+
+            Calendar hierarchies such as year, month, week etc. must be transformed into specific intervals by the clients. For each interval, all data along the dimension will be passed through the reducer. The computed values will be projected to the labels, so the number of labels and the number of intervals need to be equal.
+
+            If the dimension is not set, the data cube is expected to only have one temporal dimension.
+
+            :param intervals: Temporal left-closed intervals so that the start time is contained, but not the end time.
+            :param reducer: A reducer to be applied on all values along the specified dimension. The reducer must be a callable process (or a set processes) that accepts an array and computes a single return value of the same type as the input values, for example median.
+            :param labels: Labels for the intervals. The number of labels and the number of groups need to be equal.
+            :param dimension: The temporal dimension for aggregation. All data along the dimension will be passed through the specified reducer. If the dimension is not set, the data cube is expected to only have one temporal dimension.
+            :param context: Additional data to be passed to the reducer. Not set by default.
+
+            :return: An ImageCollection containing  a result for each time window
+        """
+        return self.process(
+            process_id="aggregate_temporal",
+            arguments=dict_no_none(
+                data= THIS,
+                intervals = intervals,
+                labels = labels,
+                dimension = dimension,
+                reducer = self._get_callback(reducer, parent_parameters=["data"])
+
+            )
+        )
+
     def ndvi(self, nir: str = None, red: str = None, target_band: str = None) -> 'DataCube':
         """ Normalized Difference Vegetation Index (NDVI)
 
@@ -677,12 +812,17 @@ class DataCube(ImageCollection):
 
             :return: a DataCube instance
         """
+        if target_band is None:
+            metadata = self.metadata.reduce_dimension(self.metadata.band_dimension.name)
+        else:
+            metadata = self.metadata.append_band(Band(target_band, "ndvi", None))
         return self.process(
             process_id='ndvi',
             arguments=dict_no_none(
                 data=THIS,
                 nir=nir, red=red, target_band=target_band
-            )
+            ),
+            metadata=metadata
         )
 
     def rename_dimension(self, source: str, target: str):
@@ -702,7 +842,8 @@ class DataCube(ImageCollection):
                 data=THIS,
                 source=self.metadata.assert_valid_dimension(source),
                 target=target
-            )
+            ),
+            metadata=self.metadata.rename_dimension(source,target)
         )
 
     def rename_labels(self, dimension: str, target: list, source: list = None) -> 'DataCube':
@@ -721,27 +862,10 @@ class DataCube(ImageCollection):
                 dimension=self.metadata.assert_valid_dimension(dimension),
                 target=target,
                 source=source
-            )
+            ),
+            metadata=self.metadata.rename_labels(dimension,target,source)
         )
 
-    @deprecated("use 'linear_scale_range' instead")
-    def stretch_colors(self, min, max) -> 'DataCube':
-        """ Color stretching
-        deprecated, use 'linear_scale_range' instead
-
-            :param min: Minimum value
-            :param max: Maximum value
-            :return: a DataCube instance
-        """
-        process_id = 'stretch_colors'
-        args = {
-            'data': THIS,
-            'min': min,
-            'max': max
-        }
-
-        return self.process(process_id, args)
-
     def linear_scale_range(self, input_min, input_max, output_min, output_max) -> 'DataCube':
         """ Color stretching
             :param input_min: Minimum input value
@@ -830,35 +954,66 @@ class DataCube(ImageCollection):
             )
         )
 
-    def merge(self, other: 'DataCube', overlap_resolver: Union[str, PGNode] = None) -> 'DataCube':
+    def merge_cubes(
+            self, other: 'DataCube', overlap_resolver: Union[str, PGNode, typing.Callable] = None
+    ) -> 'DataCube':
+        """
+        Merging two data cubes
+
+        The data cubes have to be compatible. A merge operation without overlap should be reversible with (a set of) filter operations for each of the two cubes. The process performs the join on overlapping dimensions, with the same name and type.
+        An overlapping dimension has the same name, type, reference system and resolution in both dimensions, but can have different labels. One of the dimensions can have different labels, for all other dimensions the labels must be equal. If data overlaps, the parameter overlap_resolver must be specified to resolve the overlap.
+
+        Examples for merging two data cubes:
+
+        #. Data cubes with the dimensions x, y, t and bands have the same dimension labels in x,y and t, but the labels for the dimension bands are B1 and B2 for the first cube and B3 and B4. An overlap resolver is not needed. The merged data cube has the dimensions x, y, t and bands and the dimension bands has four dimension labels: B1, B2, B3, B4.
+        #. Data cubes with the dimensions x, y, t and bands have the same dimension labels in x,y and t, but the labels for the dimension bands are B1 and B2 for the first data cube and B2 and B3 for the second. An overlap resolver is required to resolve overlap in band B2. The merged data cube has the dimensions x, y, t and bands and the dimension bands has three dimension labels: B1, B2, B3.
+        #. Data cubes with the dimensions x, y and t have the same dimension labels in x,y and t. There are two options:
+                * Keep the overlapping values separately in the merged data cube: An overlap resolver is not needed, but for each data cube you need to add a new dimension using add_dimension. The new dimensions must be equal, except that the labels for the new dimensions must differ by name. The merged data cube has the same dimensions and labels as the original data cubes, plus the dimension added with add_dimension, which has the two dimension labels after the merge.
+                * Combine the overlapping values into a single value: An overlap resolver is required to resolve the overlap for all pixels. The merged data cube has the same dimensions and labels as the original data cubes, but all pixel values have been processed by the overlap resolver.
+        #. Merging a data cube with dimensions x, y, t with another cube with dimensions x, y will join on the x, y dimension, so the lower dimension cube is merged with each time step in the higher dimensional cube. This can for instance be used to apply a digital elevation model to a spatiotemporal data cube.
+
+        @param other: The data cube to merge with.
+        @param overlap_resolver: A reduction operator that resolves the conflict if the data overlaps. The reducer must return a value of the same data type as the input values are. The reduction operator may be a single process such as multiply or consist of multiple sub-processes. null (the default) can be specified if no overlap resolver is required.
+        @return: The merged data cube.
+        """
         arguments = {
             'cube1': {'from_node': self._pg},
             'cube2': {'from_node': other._pg},
         }
         if overlap_resolver:
-            if isinstance(overlap_resolver, str):
-                # Simple resolver (specified as process_id string)
-                overlap_resolver = PGNode(
-                    process_id=overlap_resolver,
-                    arguments={"data": [{"from_parameter": "x"}, {"from_parameter": "y"}]}
-                )
-            arguments["overlap_resolver"] = {"process_graph": overlap_resolver}
+            arguments["overlap_resolver"] = self._get_callback(overlap_resolver, parent_parameters=["x", "y"])
         # TODO #125 context
         # TODO: set metadata of reduced cube?
-        return self.process_with_node(PGNode(process_id="merge_cubes", arguments=arguments))
+        return self.process(process_id="merge_cubes", arguments=arguments)
 
-    def apply_kernel(self, kernel: Union[np.ndarray, List[List[float]]], factor=1.0) -> 'DataCube':
+    merge = legacy_alias(merge_cubes, name="merge")
+
+    def apply_kernel(self, kernel: Union[np.ndarray, List[List[float]]], factor=1.0, border = 0, replace_invalid=0) -> 'DataCube':
         """
         Applies a focal operation based on a weighted kernel to each value of the specified dimensions in the data cube.
 
-        :param kernel: The kernel to be applied on the data cube. It should be a 2D (numpy) array.
+        The border parameter determines how the data is extended when the kernel overlaps with the borders.
+        The following options are available:
+
+        * numeric value - fill with a user-defined constant number n: nnnnnn|abcdefgh|nnnnnn (default, with n = 0)
+        * replicate - repeat the value from the pixel at the border: aaaaaa|abcdefgh|hhhhhh
+        * reflect - mirror/reflect from the border: fedcba|abcdefgh|hgfedc
+        * reflect_pixel - mirror/reflect from the center of the pixel at the border: gfedcb|abcdefgh|gfedcb
+        * wrap - repeat/wrap the image: cdefgh|abcdefgh|abcdef
+
+
+        :param kernel: The kernel to be applied on the data cube. The kernel has to be as many dimensions as the data cube has dimensions.
         :param factor: A factor that is multiplied to each value computed by the focal operation. This is basically a shortcut for explicitly multiplying each value by a factor afterwards, which is often required for some kernel-based algorithms such as the Gaussian blur.
+        :param border: Determines how the data is extended when the kernel overlaps with the borders. Defaults to fill the border with zeroes.
+        :param: replace_invalid: This parameter specifies the value to replace non-numerical or infinite numerical values with. By default, those values are replaced with zeroes.
         :return: A data cube with the newly computed values. The resolution, cardinality and the number of dimensions are the same as for the original data cube.
         """
         return self.process('apply_kernel', {
             'data': THIS,
             'kernel': kernel.tolist() if isinstance(kernel, np.ndarray) else kernel,
-            'factor': factor
+            'factor': factor,
+            'border': border,
+            'replace_invalid': replace_invalid
         })
 
     def raster_to_vector(self) -> 'VectorCube':
@@ -925,8 +1080,9 @@ class DataCube(ImageCollection):
 
         return self._polygonal_timeseries(polygon, "sd")
 
-    def _polygonal_timeseries(self, polygon: Union[Polygon, MultiPolygon, str], func: str) -> 'DataCube':
-
+    def _polygonal_timeseries(
+            self, polygon: Union[Polygon, MultiPolygon, str], func: Union[str, PGNode, typing.Callable]
+    ) -> 'DataCube':
         if isinstance(polygon, str):
             # polygon is a path to vector file
             # TODO this is non-standard process: check capabilities? #104 #40
@@ -940,20 +1096,20 @@ class DataCube(ImageCollection):
                 }
             }
 
-        return self.process_with_node(PGNode(
+        return self.process(
             process_id="aggregate_spatial",
             arguments={
                 "data": self._pg,
                 "geometries": geometries,
-                "reducer": {"process_graph": PGNode(
-                    process_id=func,
-                    arguments={"data": {"from_parameter": "data"}}
-                )},
+                "reducer": self._get_callback(func, parent_parameters=["data"])
                 # TODO #125 target dimension, context
             }
-        ))
+        )
 
-    def save_result(self, format: str = "GTIFF", options: dict = None):
+    def save_result(self, format: str = "GTiff", options: dict = None):
+        formats = set(self._connection.list_output_formats().keys())
+        if format.lower() not in {f.lower() for f in formats}:
+            raise ValueError("Invalid format {f!r}. Should be one of {s}".format(f=format, s=formats))
         return self.process(
             process_id="save_result",
             arguments={
@@ -963,8 +1119,17 @@ class DataCube(ImageCollection):
             }
         )
 
-    def download(self, outputfile: str, format: str = "GTIFF", options: dict = None):
-        """Download image collection, e.g. as GeoTIFF."""
+    def download(self, outputfile: Union[str, pathlib.Path, None] = None, format: str = "GTIFF", options: dict = None):
+        """
+        Download image collection, e.g. as GeoTIFF.
+        If :param: outputfile is provided, the result is stored on disk locally, otherwise, a bytes object is returned.
+        The bytes object can be passed on to a suitable decoder for decoding.
+
+        :param outputfile: Optional, an output file if the result needs to be stored on disk.
+        :param format: Optional, defaults to "GTIFF", an output format supported by the backend.
+        :param options: Optional, file format options
+        :return: None if the result is stored to disk, or a bytes object returned by the backend.
+        """
         newcollection = self.save_result(format=format, options=options)
         return self._connection.download(newcollection._pg.flatten(), outputfile)
 
@@ -973,7 +1138,7 @@ class DataCube(ImageCollection):
 
     def execute_batch(
             self,
-            outputfile: Union[str, pathlib.Path], out_format: str = None,
+            outputfile: Union[str, pathlib.Path] = None, out_format: str = None,
             print=print, max_poll_interval=60, connection_retry_interval=30,
             job_options=None, **format_options) -> RESTJob:
         """
@@ -994,7 +1159,6 @@ class DataCube(ImageCollection):
             outputfile=outputfile,
             print=print, max_poll_interval=max_poll_interval, connection_retry_interval=connection_retry_interval
         )
-
     def send_job(self, out_format=None, job_options=None, **format_options) -> RESTJob:
         """
         Sends a job to the backend and returns a Job instance. The job will still need to be started and managed explicitly.
@@ -1026,6 +1190,54 @@ class DataCube(ImageCollection):
         """Executes the process graph of the imagery. """
         return self._connection.execute(self._pg.flatten())
 
+    @staticmethod
+    def execute_local_udf(udf: str, datacube: Union[str, 'xarray.DataArray', 'openeo_udf.api.datacube.DataCube'] =None , fmt='netcdf'):
+        """
+        Locally executes an user defined function on a previously downloaded datacube.
+        
+        :param udf: the code of the user defined function
+        :param datacube: the path to the downloaded data in disk or a DataCube
+        :param fmt: format of the file if datacube is string
+        :return: the resulting DataCube
+        """
+        from openeo_udf.api.udf_data import UdfData
+        from openeo_udf.api.run_code import run_user_code
+        from openeo_udf.api.datacube import DataCube as udf_DataCube
+        from xarray import DataArray
+        
+        udf_data=None
+        # if it is a datacube
+        if datacube is not None:
+            # get input
+            if isinstance(datacube, str):
+                from openeo.rest.conversions import datacube_from_file
+                d=datacube_from_file(datacube, fmt)
+            elif isinstance(datacube, udf_DataCube):
+                d=datacube
+            elif isinstance(datacube, DataArray):
+                d=udf_DataCube(datacube)
+            else:
+                raise TypeError("Data should be either file name to the Data or a DataCube, got "+str(datacube.__class__ if datacube is not None else 'None'))
+            # datacube's data is to be float and x,y not provided
+            d=udf_DataCube(d.get_array()
+                .astype(numpy.float64)
+                .drop(labels='x')
+                .drop(labels='y')
+            )
+            # wrap to udf_data
+            udf_data=UdfData(datacube_list=[d])
+            
+        # TODO: enrich to other types like time series, vector data,... probalby by adding  named arguments
+        # signature: UdfData(proj, datacube_list, feature_collection_list, structured_data_list, ml_model_list, metadata)
+        
+        # run the udf through the same routine as it would have been parsed in the backend
+        if udf_data is not None:
+            result=run_user_code(udf, udf_data)
+            return result
+        
+        return None
+
+
     def to_graphviz(self):
         """
         Build a graphviz DiGraph from the process graph
